/*

 File: analyticcontrol.cpp, 2003/06/17

 ACETAF, Version 3.0.1

 Copyright (C) Markus Neher, markus.neher@math.uni-karlsruhe.de
               Ingo Eble,    ingoeble@web.de

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <algorithm>
#include <functional>
#include <map>

#include "adaptintval.h"

using Adapt::Interval;
using Adapt::inf;
using Adapt::sup;

#include "cinterval.h"
#include "analyticcontrol.h"

//Type APair
bool operator <  (const APair& a, const APair& b)
{
  return ( a.getFunction() == b.getFunction() && a.getDomain() < b.getDomain() );
}

bool operator <= (const APair& a, const APair& b)
{
  return ( a.getFunction() == b.getFunction() && a.getDomain() <= b.getDomain() );
}

bool operator == (const APair& a, const APair& b)
{
  return ( a.getFunction() == b.getFunction() && a.getDomain() == b.getDomain() );
}

//Type AHistory
void AHistory::add(const APair& ap)
{
  history.remove_if( std::bind2nd( std::less_equal<APair>(), ap )  );
  history.push_back( ap );
}

bool AHistory::contain(const APair& ap) const
{
  return ( history.end() != find_if( history.begin(), history.end(), std::bind1st( std::less_equal<APair>(), ap )) );
}

//Type ACheck
bool ACheck::isRare(const CInterval& z)
{
  try
    {
      cifunction( z );
      return false;
    }
  catch(division_by_zero)
    {
      return true;
    }
  catch(function_not_defined)
    {
      return true;
    }
}

CInterval ACheck::enclosure(const Circle& c)//Returns an rectangular enclosure of 'c'
{
  double tmp1,tmp2,tmp3,tmp4;

  double 
    rc = c.getCenter().real(),
    ic = c.getCenter().imag(),
    r  = c.getRadius();

#ifdef FILIB_VERSION

  SUB_DOWN( tmp1, rc, r );
  ADD_UP  ( tmp2, rc, r );
  SUB_DOWN( tmp3, ic, r );
  ADD_UP  ( tmp4, ic, r );

#else //C-XSC Version

  tmp1 = cxsc::_double(cxsc::subdown( rc, r));
  tmp2 = cxsc::_double(cxsc::addup  ( rc, r));
  tmp3 = cxsc::_double(cxsc::subdown( ic, r));
  tmp4 = cxsc::_double(cxsc::addup  ( ic, r));

#endif

  return CInterval( Interval( tmp1, tmp2 ),   //Re
		    Interval( tmp3, tmp4 )  );//Im
}

bool ACheck::isRare_in(const CInterval& z,const Complex& c,double& r)
{
  if( stopThread   ) return false;

  std::multimap<double,CInterval,std::less<double> > ListOfCIntervals;
  std::multimap<double,CInterval>::iterator actpos;

  ListOfCIntervals.insert( std::make_pair( inf( abs(z-CInterval(c)) ), z ) );

  bool go = true;
  int k = 0;
  while( go )
    {
      ++k;
      actpos = ListOfCIntervals.begin();

      if( isRare( actpos->second ) )//Function not defined
	{
	  if( abs(diam(actpos->second))/2.0 < 1e-3 )
	    {
	      r = ( (actpos->first)<r )?(actpos->first):r;
	      go = false;
	    }
	  else
	    {
	      Interval 
		re = (actpos->second).re(), 
		im = (actpos->second).im();

	      Interval 
		re1( re.inf(), re.mid() ),
		re2( re.mid(), re.sup() ),
		im1( im.inf(), im.mid() ),
		im2( im.mid(), im.sup() );

	      ListOfCIntervals.erase( actpos );

	      CInterval
		z1( re1, im1 ),
		z2( re2, im1 ),
		z3( re1, im2 ),
		z4( re2, im2 );

	      ListOfCIntervals.insert( std::make_pair( inf( abs(z1-CInterval(c)) ), z1 ) );
	      ListOfCIntervals.insert( std::make_pair( inf( abs(z2-CInterval(c)) ), z2 ) );
	      ListOfCIntervals.insert( std::make_pair( inf( abs(z3-CInterval(c)) ), z3 ) );
	      ListOfCIntervals.insert( std::make_pair( inf( abs(z4-CInterval(c)) ), z4 ) );
	    }
	}
      else 
	{
          ListOfCIntervals.erase( actpos );
	  if( k==1 || ListOfCIntervals.size() == 0 ) go = false;
	}
    }

  if( k==1 ) return false;
  else       return true;
}

void ACheck::stop()
{
  stopThread = true;
}

bool ACheck::stopped()
{
  return stopThread;
}

void ACheck::run()
{
  Circle K( res->getDomain() );

  if( !isRare( enclosure( K ) ) ) return;

  K.zoom( 0.5 );

  while( isRare( enclosure( K ) ) )
    {
      if( K.getRadius() <= 0.1 ) //when the circle is too small, get out of here!!
	{
	  res->setDomain( Circle( K.getCenter(), 0.0 ) );
	  return;
	}
      K.zoom( 0.5 );
      if( stopThread ) return;
    }

  double r,r_var,d,rc( K.getCenter().real() ),ic( K.getCenter().imag() ),sqrt_2( sup(sqrt(Interval(2.0))) );
  Complex c( K.getCenter() );

  Interval A,B,C,D,E,F;
  bool test = false, go = true;

  while( go )
    {
      r = K.getRadius(); 

#ifdef FILIB_VERSION

      MUL_UP( d, sqrt_2, r );
      r_var = d; 
      
      double tmp1,tmp2;

      SUB_DOWN( tmp1, rc, d );
      SUB_UP( tmp2, rc, r );
      A = Interval( tmp1, tmp2 );
      SUB_DOWN( tmp1, rc, r );
      ADD_UP( tmp2, rc, r );
      B = Interval( tmp1, tmp2 );
      ADD_DOWN( tmp1, rc, r );
      ADD_UP( tmp2, rc, d );
      C = Interval( tmp1, tmp2 );
      SUB_DOWN( tmp1, ic, d );
      SUB_UP( tmp2, ic, r );
      D = Interval( tmp1, tmp2 );
      SUB_DOWN( tmp1, ic, r );
      ADD_UP( tmp2, ic, r );
      E = Interval( tmp1, tmp2 );
      ADD_DOWN( tmp1, ic, r );
      ADD_UP( tmp2, ic, d );
      F = Interval( tmp1, tmp2 );

#else //C-XSC Version

      r_var = d = cxsc::_double(cxsc::multup( sqrt_2, r ));

      A = Interval( cxsc::_double(cxsc::subdown( rc, d )), cxsc::_double(cxsc::subup( rc, r )) );
      B = Interval( cxsc::_double(cxsc::subdown( rc, r )), cxsc::_double(cxsc::addup( rc, r )) );
      C = Interval( cxsc::_double(cxsc::adddown( rc, r )), cxsc::_double(cxsc::addup( rc, d )) );
      D = Interval( cxsc::_double(cxsc::subdown( ic, d )), cxsc::_double(cxsc::subup( ic, r )) );
      E = Interval( cxsc::_double(cxsc::subdown( ic, r )), cxsc::_double(cxsc::addup( ic, r )) );
      F = Interval( cxsc::_double(cxsc::adddown( ic, r )), cxsc::_double(cxsc::addup( ic, d )) );

#endif      

      if( isRare_in( CInterval( C,  E ), c, r_var ) && ( r_var < d ) )//Check right interval
	{
	  d  = r_var;
	  go = false;
	}
      if( isRare_in( CInterval( B,  F ), c, r_var ) && ( r_var < d ) )//Check upper interval
	{
	  d  = r_var;
	  go = false;
	}
      if( isRare_in( CInterval( A,  E ), c, r_var ) && ( r_var < d ) )//Check left interval
	{
	  d  = r_var;
	  go = false;
	}
      if( isRare_in( CInterval( B,  D ), c, r_var ) && ( r_var < d ) )//Check lower interval
	{
	  d  = r_var;
	  go = false;
	}

      K.setRadius( d );// d < sqrt_2 * r, if go = false, otherwise d = sqrt_2 * r.
	
      if( go )//All checks above were succesfull
	{ //Then check the interval's in the four corners
	  if( isRare( CInterval( C, D ) ) || //Left-lower, Left-upper, Right-upper, Right-lower
	      isRare( CInterval( C, F ) ) ||   
	      isRare( CInterval( A, F ) ) ||   
	      isRare( CInterval( A, D ) )    ) go = false;//Not succesfull
	}
      if( stopThread ) return;
    }

  res->setDomain( K );
}

// Type ACheckControl
ACheckControl::ACheckControl(AHistory *h,const Circle& c,const std::string& f) : history( h )
{
  ap         = new APair ( f, c );
  check      = new ACheck( ap );
}

ACheckControl::~ACheckControl()
{
  delete check;
  delete ap;
}

void ACheckControl::start()
{
  startTimer( 100 );
  check->start();
}

void ACheckControl::stop()
{
  check->stop();
}

void ACheckControl::timerEvent(QTimerEvent*)
{
  if( !(check->running()) )
    {
      killTimers();
      if( check->stopped() )
	{
	  emit checkStopped();
	  return;
	}
      else
	{
	  history->add( *ap );
	  emit checkFinished( ap );
	  return;
	}
    }
}

#include "analyticcontrol.moc"

/*

  End of File: analyticcontrol.cpp

*/
