/*

 File: mathexpr.h, 2008/02/11

 ACETAF, Version 3.0.1

 Copyright (C) Markus Neher, markus.neher@math.uni-karlsruhe.de
               Ingo Eble,    ingoeble@web.de

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifndef _EXPRESSION_H_INCLUDED
#define _EXPRESSION_H_INCLUDED

#include <vector>
#include <string>

#include "error.h"

template<class T> class prim;
template<class T> class term;

// begin insertion 080211
template<class T> class MathExpr;
template<class T> std::ostream& operator <<(std::ostream& x, const MathExpr<T>& y);
// end insertion 080211

template<class T> class MathExpr
{
  std::vector<term<T> > terms;
  std::string s;

 public:

  MathExpr() : terms(0), s("") {}
  MathExpr(const std::string&,const std::string&) throw(syntax_error);
  MathExpr(const MathExpr<T>& e) : terms( e.terms ), s( e.s ) {}
  ~MathExpr() {}

  MathExpr& operator = (const MathExpr<T>& e)
  {
    if( this == &e ) return *this;

    terms = e.terms;
    s     = e.s;
    return *this;
  }

  friend std::ostream& operator << <>(std::ostream&, const MathExpr&);
  T operator () (const T& x = T(0)) //Evaluation in one independent variable
    {
      typename std::vector<term<T> >::iterator first = terms.begin(), last = terms.end();
      T sum( (*first)(x) );
      while( ++first != last )
	{
	  sum += (*first)(x);
	}
      return sum;
    }
};

template<class T> std::ostream& operator <<(std::ostream& os, const MathExpr<T>& e) 
{
  os << e.s;
  return os;
}

template<class T> class term
{
  std::vector<prim<T> > prims;

 public:

  term() : prims(0) {}
  term(const std::string&,const std::string&) throw(syntax_error);
  term(const term<T>& t) : prims( t.prims ) {}
  ~term() {}

  term& operator = (const term<T>& t)
  {
    if( this == &t ) return *this;

    prims = t.prims;
    return *this;
  }
  
  T operator () (const T& x) 
    {
      typename std::vector<prim<T> >::iterator first = prims.begin(), last = prims.end();
      T prod( (*first)(x) );
      while( ++first != last )
	{
	  prod *= (*first)(x);
	}
      return prod;
    }
};

template<class T> class function_T
{
  typedef T (*pfunc)(const T&);
  
  pfunc pf;
  MathExpr<T> *arg;

 public:

  function_T() : pf(0), arg(0) {}
  function_T(pfunc p) : pf(p), arg(0) {}
  function_T(const function_T& ft) : pf(ft.pf)
    { 
      if( ft.arg ) arg = new MathExpr<T>( *(ft.arg) ); 
      else         arg = 0;
    }
  ~function_T() 
    { 
      if( arg ) delete arg;
    }

  void set_arg(const MathExpr<T>& e) 
    { 
      if( arg ) delete arg;
      arg = new MathExpr<T>( e ); 
    }
  T operator () (const T& x) { return (*pf)( (*arg)( x ) ); }
};

template<class T> class function_TT
{
  typedef T (*pfunc)(const T&,const T&);
  
  pfunc pf;
  MathExpr<T> *arg1, *arg2;

  public:

  function_TT() : pf(0), arg1(0), arg2(0) {}
  function_TT(pfunc p) : pf(p), arg1(0), arg2(0) {}
  function_TT(const function_TT& ftt) : pf(ftt.pf)
    { 
      if( ftt.arg1 ) arg1 = new MathExpr<T>( *(ftt.arg1) );
      else           arg1 = 0;
      if( ftt.arg2 ) arg2 = new MathExpr<T>( *(ftt.arg2) );
      else           arg2 = 0;
    }
  ~function_TT() 
    {
      if( arg1 ) delete arg1;
      if( arg2 ) delete arg2;
    }

  void set_1st_arg(const MathExpr<T>& e) 
    { 
      if( arg1 ) delete arg1;
      arg1 = new MathExpr<T>( e ); 
    }
  void set_2nd_arg(const MathExpr<T>& e) 
    { 
      if( arg2 ) delete arg2;
      arg2 = new MathExpr<T>( e ); 
    }
  T operator () (const T& x) { return (*pf)( (*arg1)( x ), (*arg2)( x )); }
};

template<class T> class function_Ti
{
  typedef T (*pfunc)(const T&,int);
  
  pfunc pf;
  MathExpr<T>   *arg1; 
  MathExpr<int> *arg2;

  public:

  function_Ti() : pf(0), arg1(0), arg2(0) {}
  function_Ti(pfunc p) : pf(p), arg1(0), arg2(0) {}
  function_Ti(const function_Ti& fti) : pf(fti.pf)
    { 
      if( fti.arg1 ) arg1 = new MathExpr<T>( *(fti.arg1) );
      else           arg1 = 0;
      if( fti.arg2 ) arg2 = new MathExpr<int>( *(fti.arg2) );
      else           arg2 = 0;
    }
  ~function_Ti() 
    {
      if( arg1 ) delete arg1;
      if( arg2 ) delete arg2;
    }

  void set_1st_arg(const MathExpr<T>& e)   
    { 
      if( arg1 ) delete arg1;
      arg1 = new MathExpr<T>( e ); 
    }
  void set_2nd_arg(const MathExpr<int>& e) 
    { 
      if( arg2 ) delete arg2;
      arg2 = new MathExpr<int>( e ); 
    }
  T operator () (const T& x) { return (*pf)( (*arg1)( x ), (*arg2)() ); }
};

#include "zero.h"

template<class T> class prim
{
  function_T<T>  *fp1; //Funktion like "sin", "cos", etc.
  function_Ti<T> *fp2; //Funktion like "power", "root", etc.
  function_TT<T> *fp3; //Funktion like "pow"
  MathExpr<T> *expr; //Expression, e.g. the argument of a function.
  T *return_value;     //Constant value like "12", "-2.341", etc.  
  
  bool VZ; //'+' or '-'
  bool OP; //'*' or '/'

 public:

  prim() : fp1(0), fp2(0), fp3(0), expr(0), return_value(0), VZ(false), OP(false) {}  
  prim(const std::string&,const std::string&) throw(syntax_error);
  prim(const prim<T>& p) : VZ( p.VZ ), OP( p.OP )
    {
      if( p.fp1          ) fp1 = new function_T<T> ( *(p.fp1) );
      else                 fp1 = 0;
      if( p.fp2          ) fp2 = new function_Ti<T>( *(p.fp2) );
      else                 fp2 = 0;
      if( p.fp3          ) fp3 = new function_TT<T>( *(p.fp3) );
      else                 fp3 = 0;
      if( p.expr         ) expr = new MathExpr<T>( *(p.expr) );
      else                 expr = 0;
      if( p.return_value ) return_value = new T( *(p.return_value) );
      else                 return_value = 0;
    }
  ~prim() 
    {
      if( fp1          ) delete fp1;
      if( fp2          ) delete fp2;
      if( fp3          ) delete fp3;
      if( expr         ) delete expr;
      if( return_value ) delete return_value;
    }

  prim& operator = (const prim<T>& p)
  { 
    if( this == &p ) return *this;

    if( fp1          ) delete fp1;
    if( fp2          ) delete fp2;
    if( fp3          ) delete fp3;
    if( expr         ) delete expr;
    if( return_value ) delete return_value;

    if( p.fp1          ) fp1 = new function_T<T> ( *(p.fp1) );
    else                 fp1 = 0;
    if( p.fp2          ) fp2 = new function_Ti<T>( *(p.fp2) );
    else                 fp2 = 0;
    if( p.fp3          ) fp3 = new function_TT<T>( *(p.fp3) );
    else                 fp3 = 0;
    if( p.expr         ) expr = new MathExpr<T>( *(p.expr) );
    else                 expr = 0;
    if( p.return_value ) return_value = new T( *(p.return_value) );
    else                 return_value = 0;

    return *this;
  }

  T operator () (const T& x)
    {
      if( fp1 == 0 )
	{
	  if( fp2 == 0 )
	    {
	      if( fp3 == 0 )
		{
		  if( expr == 0 )
		    {
		      if( return_value == 0 )
			{
			  if( OP && is_zero<T>()( x ) ) throw division_by_zero();
			  return OP?(VZ?T(-1.0)/x:T(1.0)/x):(VZ?-x:x);
			}
		    }
		  else
		    {
		      return_value = new T( (*expr)(x) );
		    }
		}
	      else
		{
		  return_value = new T( (*fp3)( x ) );
		}
	    }
	  else
	    {
	      return_value = new T( (*fp2)( x ) );
	    }
	}
      else
	{
	  return_value = new T( (*fp1)( x ) );
	}

      if( OP && is_zero<T>()( *return_value ) ) throw division_by_zero();

      if( OP )
	{
	  if( VZ ) return T(-1) / (*return_value);
	  else     return T( 1) / (*return_value);
	}
      else
	{
	  if( VZ ) return -(*return_value);
	  else     return  (*return_value);
	}
    }
};
    
#include "mathexpr.tpl"

#endif

/*

  End of File: mathexpr.h

*/
