/*

 File: methods.cpp, 2003/06/17

 ACETAF, Version 3.0.1

 Copyright (C) Markus Neher, markus.neher@math.uni-karlsruhe.de
               Ingo Eble,    ingoeble@web.de

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "methods.h"

using Adapt::PI;
using Adapt::sup;
using Adapt::inf;

#include "cinterval.h"  //for type complex-interval

const Complex unit_i( 0.0 , 1.0 );

/*-----------------------------+
 | Compute Taylor coefficients |
 +-----------------------------*/
void MethodTwoAndFour::calculateTcoeff(int ord)
{
  tcoeff = citaylfunction( CITaylor( CInterval( domain.getCenter() ), ord ) );
}

void MethodFour::calculateProducts()
{
  products.resize( polyord+1 );
  for(int k = 1; k <= polyord+1; k++)
    {
      products[k-1] = 1.0;
      for(int n = 1; n <= deriveord; n++)
	{
	  products[k-1] *= ( (k-1) + n );
	}
    }
}

/*---------------------------------------------------+
 | In Method 1, a complex interval x is built from   |
 | "phi", "center" and "radius"                      |
 | f(x) is enclosed, abs( f(x) ) is returned         |
 +---------------------------------------------------*/
void MethodOne::operator () (const Interval& phi, Interval& f_value, Interval& f_mid)
{
  //Mean value form evaluation of f

  /************* Evaluate function **************************/
  
  //Cover segment by rectangular complex interval
  CInterval x( domain.getCenter().real() + domain.getRadius() * cos( phi ),   //Re
	       domain.getCenter().imag() + domain.getRadius() * sin( phi ) ); //Im
 
  //Enclose midpoint
  CInterval midx( x.mid() );

  //Evaluate f
  const CInterval fmid = cifunction( midx );
  CITaylor xx( x, 1 );
  const CITaylor df = citaylfunction( xx );

  //Return abs( intersection of standard evaluation and mean value form evaluation of f) 

  f_mid   = abs( fmid );
  f_value = abs( (df).derivative( 0 ) & ( fmid + ( x - midx ) * (df).derivative( 1 ) ) ); 
}
  
/*--------------------------------------------------------------------------+
 | In Method 2, the Taylor polynomial of order "torder"                     |
 | is subtracted from f.                                                    |
 | Enclosures for both f and p are computed. | f - p |_infinity is returned |
 +--------------------------------------------------------------------------*/
void MethodTwo::operator () (const Interval& phi, Interval& f_value, Interval& f_mid)
{ 
  // Centered form evaluation of parameterized function
  // Efficient centered form evaluation of p' via Horner's scheme

  /************* Evaluate function **************************/

  //Cover segment by rectangular complex interval
  CInterval x( domain.getCenter().real() + domain.getRadius() * cos( phi ),   //Re
	       domain.getCenter().imag() + domain.getRadius() * sin( phi ) ); //Im
  
  //Enclose midpoint
  CInterval midx( x.mid() );

  //Evaluate f
  const CInterval fmid = cifunction( midx );
  CITaylor xx( x, 1 );
  const CITaylor df = citaylfunction( xx );

  /************* Evaluate Taylor polynomial ************************/

  /*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    +  This works if polyord >= 1 is true, otherwise the program crashes  +
    +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

  CInterval pmid( tcoeff[ polyord ] );    
  CInterval dpmid( polyord * pmid );    
  CInterval cidp( dpmid ); 

  //Evaluate p at the midpoint
  for(int k = polyord-1; k >= 0; k--)
    {
      pmid = pmid * midx + tcoeff[k];
    }

  //Centered form evaluation of p' = dp
  for(int k= polyord-1; k >= 2; k--)
    {
      dpmid = dpmid * midx + k * tcoeff[k];
      cidp *= x;
      cidp += dpmid;
    }

  dpmid = dpmid * midx + tcoeff[1];
  CInterval dpcent( cidp * ( x - midx) + dpmid ); 

  //Return centered form evaluation of | f - p |

  f_mid   = abs( fmid - pmid );
  f_value = abs( fmid - pmid + ( x - midx ) * ( (df).derivative( 1 )  - dpcent) ); 
}

/*------------------------------------------------------------------+
 | In Method 3, the "deriveord"-th derivative of f is enclosed      |
 +------------------------------------------------------------------*/
void MethodThree::operator () (const Interval& phi, Interval& f_value, Interval& f_mid)
{
  // Mean value form evaluation of derivative

  /************* Evaluate function **************************/

  //Cover segment by rectangular complex interval
  CInterval x( domain.getCenter().real() + domain.getRadius() * cos( phi ),   //Re
	       domain.getCenter().imag() + domain.getRadius() * sin( phi ) ); //Im
 
  //Enclose midpoint
  CInterval midx( x.mid() );

  //Evaluate df
  CITaylor midxx( midx, deriveord );
  const CITaylor dfmid = citaylfunction( midxx );
  CITaylor xx( x, deriveord+1 );
  const CITaylor df = citaylfunction( xx ); 
 
  //Return abs( intersection of standard evaluation of df and mean value form evaluation of df )

  f_mid   = abs( (dfmid).derivative( deriveord ) );
  f_value = abs(  (df).derivative( deriveord ) & 
		( (dfmid).derivative( deriveord ) + ( x - midx ) * (df).derivative( deriveord + 1 ) ) ); 
}

/*--------------------------------------------------+
 | Method 4 is a combination of the Methods 2 and 3 |
 +-------------------------------------------------*/
void MethodFour::operator () (const Interval& phi, Interval& f_value, Interval& f_mid)
{
  // Mean value form evaluation of derivative

  /************* Evaluate function **************************/

  //Cover segment by rectangular complex interval
  CInterval x( domain.getCenter().real() + domain.getRadius() * cos( phi ),   //Re
	       domain.getCenter().imag() + domain.getRadius() * sin( phi ) ); //Im

  //Enclose midpoint
  CInterval midx( x.mid() );

  //Evaluate df
  CITaylor midxx( midx, deriveord );
  const CITaylor dfmid = citaylfunction( midxx );
  CITaylor xx( x, deriveord+1 );
  const CITaylor df = citaylfunction( xx ); 
 
  /************* Evaluate Taylor polynomial ************************/

  /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    +  This works if polyord >= 1 is true, otherwise the program crashes +
    ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
  CInterval pmid( products[polyord] * tcoeff[ polyord + deriveord ] );    
  CInterval dpmid( ( polyord ) * pmid );    
  CInterval cidp( dpmid ); 
 
  //Evaluate p at the midpoint
  for(int k = polyord-1; k >= 0; k--)
    { 
      pmid = pmid * midx + products[k] * tcoeff[k + deriveord];
    }

  //Centered form evaluation of p' = dp
  for(int k= polyord-1; k >= 2; k--)
    {
      dpmid = dpmid * midx + k * products[k] * tcoeff[k + deriveord];
      cidp  = cidp * x + dpmid;
    }

  dpmid = dpmid * midx + products[1] * tcoeff[1 + deriveord];

  CInterval dpcent( cidp * ( x - midx) + dpmid ); 

  //Return centered form evaluation of | df - p | 
  f_mid   = abs( (dfmid).derivative( deriveord ) - pmid );
  f_value = abs(  (dfmid).derivative( deriveord ) - pmid  
		  + ( x - midx ) * ( (df).derivative( 1 + deriveord ) - dpcent ) );
}

//----------------------------------------

#include "stringutil.h"
#include "error.h"

void OptTaylOrd::cauchytest(double& max,double& mid_max)
{
  //twoPi = [ 2*pi, 2*pi ]
  static Interval Two_PI = 2.0 * PI();
  double argdiam;

#ifdef FILIB_VERSION
  DIV_UP( argdiam, sup(Two_PI), dat->maxSegments );
#else //C-XSC Version
  argdiam = cxsc::_double( cxsc::divup( sup(Two_PI), dat->maxSegments ) );
#endif

  max = mid_max = 0.0;

  Interval h( Two_PI / dat->segments );//Steplength

  /*-------------------------+
   | Screen all subintervals |
   | Update max and mid_max  |
   +-------------------------*/

  Interval arg, Wf_arg, Wf_mid_arg;
  for(int k=1; k <= dat->segments; k++) 
    {
      if( shd->readStatus() == Thread::STOP ) return;

      arg = Interval( (k-1)* h ) + Interval( 0.0, argdiam );

      (*mth)( arg, Wf_arg, Wf_mid_arg );//Call to Method
      
      mid_max = ( mid_max < inf( Wf_mid_arg ) ) ? inf( Wf_mid_arg ) : mid_max;
      max     = ( max     < sup( Wf_arg     ) ) ? sup( Wf_arg     ) : max; 
    }
}     

#include <utility>
#include <list>

void OptTaylOrd::run()
{
  try
    {  
      int maxord = dat->maxOrder;
      double optTime;
      double max = 0.0, midmax = 1.0;

      if( mth->getPolyord() > maxord ) mth->changePolyord( maxord );
      *(res->tm) = mth->Tcoeff();

      //
      //  optimization strategy, loop I: increase order until divergence 
      //

      std::list<std::pair<int,double> > OrdMaxList( maxord/5 + 1 );
      std::list<std::pair<int,double> >::iterator first = OrdMaxList.begin(), actual = first--;
      double oldmax( 0.0 ), oldmidmax( 0.0 ), optMax( 0.0 ); //!!!!!!!!!!!!!!!!!!!!!!!!

      int Ord = mth->getPolyord()-5, optOrd = 0;

      //maxord +=1; ???????????
      while( max < 100*midmax && Ord <= maxord-5 ) // <=> !( max >= 100*midmax ) && Ord <= maxord-5
	{ 
	  Ord += 5; //current Order
	  mth->changePolyord( Ord ); //set current order for calculation

	  QString mess( "Current order = " );
	  mess += int2QStr( Ord, "%4i" );

	  CPU_time.start();
	  cauchytest( max, midmax ); 
	  CPU_time.stop();

	  mess += "   |f-T_l| = " + double2QStr( midmax, "% 9.8lE" )
	       +  "   |F-T_l| = " + double2QStr( max   , "% 9.8lE" );

	  shd->appendToMessage( mess );

	  if( shd->readStatus() == Thread::STOP ) return;

          if( optOrd == 0 ) // first time
	    {
	      optMax = max; 
	      optOrd = Ord;
	    }

	  if( max < optMax )
	    {
	      (*actual++) = std::make_pair( optOrd, optMax );
	      optMax = max;
	      optOrd = Ord;
	    } 
	}
      
      // save calculated taylorcoefficients
      *(res->tm) = mth->Tcoeff();

      double twopercentmark = 1.02 * optMax;
      int minord = 0;
            
      --actual;
      while( actual != first )
	{
	  if( actual->second > twopercentmark )
	    {
	      minord = actual->first;
	      actual = first;
	    }
	  else --actual;
	}

      //
      //  optimization strategy, loop II:
      //  decrease order until backward divergence 
      //   

      maxord = optOrd;
            
      Ord = (minord + maxord)/2;      
      while( Ord != minord )
	{
	  mth->changePolyord( Ord ); //set current order

	  QString mess( "Current order = " );
	  mess += int2QStr( Ord, "%4i" );

	  CPU_time.start();
	  cauchytest( max, midmax ); 
	  CPU_time.stop();

	  mess += "   |f-T_l| = " + double2QStr( midmax, "% 9.8lE" )
	       +  "   |F-T_l| = " + double2QStr( max   , "% 9.8lE" );

	  shd->appendToMessage( mess );

	  if( shd->readStatus() == Thread::STOP ) return;
	  
	  if( max < optMax )
            {
	      optMax = max; 
	      optOrd = Ord;
	      twopercentmark = 1.02*max;
	    }

	  if( max <= twopercentmark ) maxord = Ord;
	  else                        minord = Ord;
	  Ord = (minord + maxord)/2;      
	}

      ++Ord;

      res->optTime = CPU_time.getElapsedTime();

      //  set 0.95-optimal order
      mth->changePolyord( Ord );

      QString mess( "\nOptimal order = " );
      mess += int2QStr( optOrd, "%4i" );
      mess += "\n\nModified Cauchy's estimate will be computed for 0.98-optimal order ";
      mess += QString::number( Ord );
      mess += ".\n";
      shd->appendToMessage( mess );
    }
  catch(division_by_zero)
    {
      QString mess("Division by zero occured in the calculation.");
      shd->setError( mess );
    }
  catch(function_not_defined)
    {
      QString mess("f is not analytic on the complete domain given.");
      shd->setError( mess );
    }
  catch(...)
    {
      QString mess("An unknown error occured in the calculation.");
      shd->setError( mess );
    }
}

/*

  End of File: methods.cpp

*/
