/*

 File: methods.h, 2003/06/17

 ACETAF, Version 3.0.1

 Copyright (C) Markus Neher, markus.neher@math.uni-karlsruhe.de
               Ingo Eble,    ingoeble@web.de

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifndef _METHODS_H_INCLUDED
#define _METHODS_H_INCLUDED

#include "circle.h"
#include "methodbase.h"
#include "mathexpr.h"
#include "functions.h"
#include "taylor.h"

#include <vector>

enum Method { 
  No_Method = -1,
  Method1   =  1, 
  Method2   =  2, 
  Method3   =  3, 
  Method4   =  4 
};

//Method II and IV
class MethodTwoAndFour : public MethodBase
{
 protected:
  
  CITaylor tcoeff;

  int polyord;
  int maxpolyord;
  int coeffprec;

  MathExpr<CInterval> cifunction;
  MathExpr<CITaylor>  citaylfunction;

  void calculateTcoeff(int);

  Circle domain;

 public:

  MethodTwoAndFour(const Circle& d, //Domain
		   int ord,         //Order of taylorpolynomial
		   int prec,        //Precision of coefficients
		   const std::string& functionString) : domain( d ), polyord( ord  ), maxpolyord( ord ),
                                                        coeffprec( prec ),
                                                        cifunction    ( functionString, "z" ),
                                                        citaylfunction( functionString, "z" )
    {
      calculateTcoeff( polyord );
    }
  virtual ~MethodTwoAndFour() {}

  virtual void changePolyord(int) = 0;

  int       getPolyord() const { return polyord; }

  const CITaylor& Tcoeff () const { return tcoeff;  }
};

/*--------------------------------------------------------------------------+
 | In Method 2, the Taylor polynomial of order "ord"                        |
 | is subtracted from f.                                                    |
 | Enclosures for both f and p are computed. | f - p |_infinity is returned |
 +--------------------------------------------------------------------------*/
class MethodTwo : public MethodTwoAndFour
{
 public:

  MethodTwo(const Circle& d,               //Domain
	    int ord,                       //Order of taylorpolynomial
	    int prec,                      //Precision of coefficients
	    const std::string& functionString ) 
    : MethodTwoAndFour( d, ord, prec, functionString ) {}
  ~MethodTwo() {}

  void changePolyord(int newOrder)
    {
      polyord = newOrder;
      if( newOrder > maxpolyord )
	{
	  maxpolyord = newOrder;
	  calculateTcoeff( polyord );
	}
    }  

  int identity() const { return Method( 2 ); }
  void operator () (const Interval&,Interval&,Interval&);
};

/*-------------------------------------------------------------+
 | Method 4 is a combination of Methods 2 and 3              |
 +-------------------------------------------------------------*/
class MethodFour : public MethodTwoAndFour
{
  std::vector<Interval> products;//contains products P(j,derivOrd)  

  int deriveord;

  void calculateProducts();
  
 public:
 
  MethodFour(const Circle& d,                 //Domain
	     int ord,                         //Order of taylorpolynomial
	     int prec,                        //Precision of coefficients
	     const std::string& functionString, 
	     int dord               )         //Order of derivative
    : MethodTwoAndFour( d, ord+dord, prec, functionString ), deriveord( dord ) 
    {
      polyord -= dord;
      maxpolyord = polyord;
      calculateProducts();
    }
  ~MethodFour() {}

  void changePolyord(int newOrder)
    {
      polyord = newOrder;
      if( newOrder > maxpolyord )
	{
	  maxpolyord = newOrder;
	  calculateTcoeff( polyord + deriveord );
	  calculateProducts();
	}
    }

  int identity() const { return Method( 4 ); }
  void operator () (const Interval&,Interval&,Interval&);
};

//Method 1 and 3

/*---------------------------------------------------+
 | In Method 1, a complex interval x is built from   |
 | "phi", "center" and "radius"                      |
 | f(x) is enclosed, abs( f(x) ) is returned         |
 +---------------------------------------------------*/
class MethodOne : public MethodBase
{
  Circle domain;

  MathExpr<CInterval> cifunction;
  MathExpr<CITaylor>  citaylfunction;  
  
 public:

  MethodOne(const Circle& d, 
	    const std::string& functionString) : domain( d ),
                                                 cifunction    ( functionString, "z" ),
                                                 citaylfunction( functionString, "z" )
    {}
  ~MethodOne() {}
  
  int identity() const { return Method( 1 ); }
  void operator () (const Interval&,Interval&,Interval&);
};

/*--------------------------------------------------------------+
 | In Method 3, the "derivord"-th derivative of f is enclosed   |
 +--------------------------------------------------------------*/
class MethodThree : public MethodBase
{
  Circle domain;
  int deriveord;

  MathExpr<CInterval> cifunction;
  MathExpr<CITaylor>  citaylfunction;  

 public:

  MethodThree(const Circle& c, 
	      int d,
	      const std::string& functionString) : domain( c ), deriveord( d ),
                                                   cifunction    ( functionString, "z" ),
                                                   citaylfunction( functionString, "z" )
 {}
  ~MethodThree() {}

  int identity() const { return Method( 3 ); }
  void operator () (const Interval&,Interval&,Interval&);
};

//---------------------------------------------------------

#include <qthread.h>

#include "stopwatch.h"
#include "shareddata.h"

struct OptTaylData
{
  int segments;
  int maxSegments;
  int maxOrder;
  bool optimization;

  OptTaylData(int s, int maS, int maO, bool opt) 
    : segments( s ), maxSegments( maS ), maxOrder( maO ), optimization( opt ) 
  {}
};

struct OptTaylResult
{
  double optTime;
  TaylorMemory<CInterval> *tm;
};

class OptTaylOrd : public QThread
{
 public:

  OptTaylOrd(OptTaylData *d, OptTaylResult *r, MethodTwoAndFour *m24, SharedData *s) : dat( d   ), 
                                                                                       res( r   ), 
                                                                                       mth( m24 ), 
                                                                                       shd( s   ) 
    {};
  ~OptTaylOrd() {}

 protected:

  void run();

 private:

  stopwatch CPU_time;

  OptTaylData      *dat;
  OptTaylResult    *res;
  MethodTwoAndFour *mth;
  SharedData       *shd;

  void cauchytest(double&,double&);
};


#endif

/*

  End of File: methods.h

*/
