/*

 File: optim.cpp, 2003/07/08

 ACETAF, Version 3.0.1

 Copyright (C) Markus Neher, markus.neher@math.uni-karlsruhe.de
               Ingo Eble,    ingoeble@web.de

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "optim.h"
#include "stringutil.h"
#include "error.h"

using Adapt::PI;
using Adapt::sup;
using Adapt::inf;
using Adapt::mid;

#include <cstdio>
#include <vector>

struct Candidate                     
{                                    
  Interval y;
  double Sup_Wf_y;        
                                        
  Candidate() {}
  Candidate(const double& y_i, const double& y_s) throw() : y( y_i, y_s ) {} 
  Candidate(const Candidate& k) throw() : y( k.y ), Sup_Wf_y( k.Sup_Wf_y ) {}
  Candidate& operator = (const Candidate& c)
  {
    y = c.y;
    Sup_Wf_y = c.Sup_Wf_y;

    return *this;
  }
};

#include <algorithm>
#include <functional>

class less_infMax : public std::unary_function<double,bool>
{
public:
  
  less_infMax(const double& v) : value( v ) {}
  bool operator () (const std::pair<double, Interval>& rc) { return rc.first < value; }
  
private:
  
  double value;
};

void OptimAlgA::timerEvent(QTimerEvent* e)
{
  QObject::killTimer( e->timerId() );
  shd->setStatus( Thread::TIMEOVER );
  shd->appendToMessage( "\nPROGRAM ABORTED: Maximal computation time exceeded.\n" );

  if( !result_available )
    shd->appendToMessage( "No results could be computed. Please increase maximal computation time.\n" );
  else
    shd->appendToMessage( "The following coarse bounds have been computed:\n" );
}

void OptimAlgA::collect_results()
{
  (ores->candidates) = ListOfCandidates.size();
  (ores->max) = Interval( infMax, supMax );
  (ores->calctime) = CPU_Time.getElapsedTime();

  if( ores->relprec >= odat->relPrec ) ores->err = PREC_ERROR;
  else                                 ores->err = NO_ERROR;
}

void OptimAlgA::run()
{
  try
    {
      if( odat->maxTime != 0.0 ) QObject::startTimer( int( odat->maxTime )*1000  );      

      CPU_Time.start();
      
      //X = [ 0, 2*pi ]
      Interval X( 0.0, sup( 2*PI() ) );
      
      double& subtlety = ores->subtlety; subtlety = diam(X) / int(odat->minSegments);
      
      //Filling the list of candiates
      double 
	iX( inf(X) ), 
	sX( sup(X) );
      
      ores->segments_sum = ores->candidates = odat->minSegments;
      
      Interval Wf_y, Wf_mid_y;      
      double Sup_Wf_y,Inf_Wf_mid_y;
      for(int k = 1; k <= ores->candidates; k++) 
	{
	  Interval y( iX + (k-1) * subtlety , iX + k * subtlety );

	  //Last subinterval
	  if( k == ores->candidates ) y = Interval( iX + (int(odat->minSegments) - 1) * subtlety , sX );
	  
	  (*(odat->mth))( y, Wf_y, Wf_mid_y );
	  
	  Sup_Wf_y     = sup( Wf_y     );
	  Inf_Wf_mid_y = inf( Wf_mid_y );
	  
	  infMax = ( infMax < Inf_Wf_mid_y ) ? Inf_Wf_mid_y : infMax;
	  supMax = ( supMax < Sup_Wf_y     ) ? Sup_Wf_y     : supMax; 

	  ListOfCandidates.insert( std::make_pair( Sup_Wf_y , y ) );

	  if( shd->readStatus() == Thread::STOP || shd->readStatus() == Thread::TIMEOVER ) return;
	}

      result_available = true;
      int bisect(0);

      if( (supMax == 0.0)&&(infMax == 0.0) )
	{
	  ores->relprec = 0;
	  ores->absprec = 0;
	}
      else
	{
	  double tmp1,tmp2;
#ifdef FILIB_VERSION
	  SUB_UP( tmp1, supMax, infMax );
	  //calculate relative error
	  ADD_DOWN( tmp2, std::abs( supMax ), std::abs( infMax ) );
	  DIV_UP( ores->relprec, tmp1, tmp2 );
	  //calculate absolute error
	  MUL_UP( ores->absprec, 0.5, tmp1 );
#else //C-XSC Version
	  tmp1 = cxsc::_double( cxsc::subup( supMax, infMax ) );
	  //calculate relative error
	  tmp2 = cxsc::_double( cxsc::adddown( std::abs( supMax ), std::abs( infMax ) ) );
	  ores->relprec = cxsc::_double( cxsc::divup( tmp1, tmp2 ) ); 
	  //calculate absolute error
	  ores->absprec = cxsc::_double( cxsc::multup( 0.5, tmp1 ) );
#endif
	}
     
      if( garbageflag )
	{
	  //Garbage output
	  QString tmp( "No. of bisections: "
		       + QString::number(bisect)
		       + "\nFineness: "
		       + double2QStr( subtlety, "%.10lG" )
		       + "\nLength of list: "
		       + QString::number( ores->candidates )
		       + " elements\nEnclosure:"
		       + interval2QStr( Interval( infMax, supMax ), 10 )
		       + "\n\n" );
	  
	  shd->appendToGarbage( tmp );
	}
      
      std::multimap<double,Interval>::iterator actpos, lastpos;

      while( ( ores->relprec         >= odat->relPrec  )   &&
	     ( ores->absprec         >= odat->absPrec  )   &&
	     ( ores->segments_sum <  odat->maxSegments )       ) 
	{ 
	  actpos = ListOfCandidates.begin();

	  // Bisect 
	  double my( mid( actpos->second ) );

	  Interval y1( inf( actpos->second ) , my         );
	  Interval y2( my         , sup( actpos->second ) );

	  // Update subtlety and supMax
	  double subtlety_new;
#ifdef FILIB_VERSION
	  DIV_UP( subtlety_new, diam( actpos->second ), 2.0 );
#else //C-XSC Version
	  subtlety_new = cxsc::_double( cxsc::divup( diam( actpos->second ), 2.0 ) );
#endif
	  subtlety = (subtlety < subtlety_new) ? subtlety : subtlety_new;

	  supMax = ( actpos->first < supMax ) ? actpos->first : supMax;

	  // Now remove best element
	  ListOfCandidates.erase( actpos );

	  ++bisect;

	  // Enclose function values
	  (*(odat->mth))( y1, Wf_y, Wf_mid_y );
	      
	  Sup_Wf_y     = sup( Wf_y     );
	  Inf_Wf_mid_y = inf( Wf_mid_y );
	      
	  infMax = ( infMax < Inf_Wf_mid_y ) ? Inf_Wf_mid_y : infMax;

	  ListOfCandidates.insert( std::make_pair( Sup_Wf_y, y1 ) );


	  (*(odat->mth))( y2, Wf_y, Wf_mid_y );
	      
	  Sup_Wf_y     = sup( Wf_y     );
	  Inf_Wf_mid_y = inf( Wf_mid_y );
	      
	  infMax = ( infMax < Inf_Wf_mid_y ) ? Inf_Wf_mid_y : infMax;

	  ListOfCandidates.insert( std::make_pair( Sup_Wf_y, y2 ) );

	  (ores->segments_sum) += 1;

	  if( (supMax == 0.0)&&(infMax == 0.0) )
	    {
	      ores->relprec = 0;
	      ores->absprec = 0;
	    }
	  else
	    {
	      double tmp1,tmp2;
#ifdef FILIB_VERSION
	      SUB_UP( tmp1, supMax, infMax );
	      //calculate relative error
	      ADD_DOWN( tmp2, std::abs( supMax ), std::abs( infMax ) );
	      DIV_UP( ores->relprec, tmp1, tmp2 );
	      //calculate absolute error
	      MUL_UP( ores->absprec, 0.5, tmp1 );
#else //C-XSC Version
	      tmp1 = cxsc::_double( cxsc::subup( supMax, infMax ) );
	      //calculate relative error
	      tmp2 = cxsc::_double( cxsc::adddown( std::abs( supMax ), std::abs( infMax ) ) );
	      ores->relprec = cxsc::_double( cxsc::divup( tmp1, tmp2 ) ); 
	      //calculate absolute error
	      ores->absprec = cxsc::_double( cxsc::multup( 0.5, tmp1 ) );
#endif
	    }
	  
	  if( garbageflag )
	    {
	      //Garbage output
	      QString tmp( "No. of bisections: "
			   + QString::number(bisect)
			   + "\nFineness: "
			   + double2QStr( subtlety, "%.10lG" )
			   + "\nLength of list: "
			   + QString::number( ores->segments_sum )
			   + " elements\nEnclosure: "
			   + interval2QStr( Interval( infMax, supMax ), 10 )
			   + "\n\n" );
	  
	      shd->appendToGarbage( tmp );
	    }
	  
	  if( shd->readStatus() == Thread::STOP ) return;
	  if( shd->readStatus() == Thread::TIMEOVER ) 
	    {
	      collect_results();
	      return;
	    }
	}
      
      CPU_Time.stop();

      lastpos  = ListOfCandidates.end();
      actpos   = find_if( ListOfCandidates.begin(), lastpos, less_infMax( infMax ) );
      ListOfCandidates.erase( actpos, lastpos );
      
      /*-----------------+
       | Collect results |
       +-----------------*/
      collect_results();
    }
  catch(division_by_zero)
    {
      QString mess("Division by zero occured in the calculation.");
      shd->setError( mess );
    }
  catch(function_not_defined)
    {
      QString mess("f is not analytic on the complete domain given.");
      shd->setError( mess );
    }
  catch(...)
    {
      QString mess("An unknown error occured in the calculation.");
      shd->setError( mess );
    }
}

void OptimAlgB::timerEvent(QTimerEvent* e)
{
  QObject::killTimer( e->timerId() );
  shd->setStatus( Thread::TIMEOVER );
  shd->appendToMessage( "\nPROGRAM ABORTED: Maximal computation time exceeded.\n" );

  if( !result_available )
    shd->appendToMessage( "No results could be computed. Please increase maximal computation time.\n" );
  else
    shd->appendToMessage( "The following coarse bounds have been computed:\n" );

}

void OptimAlgB::collect_results()
{
  ores->calctime = CPU_Time.getElapsedTime();
  if( (supMax == 0.0)&&(infMax == 0.0) )
    {
      ores->relprec = 0;
      ores->absprec = 0;
    }
  else
    {
      double tmp1,tmp2;
#ifdef FILIB_VERSION
      SUB_UP( tmp1, supMax, infMax );
      //calculate relative error
      ADD_DOWN( tmp2, std::abs( supMax ), std::abs( infMax ) );
      DIV_UP( ores->relprec, tmp1, tmp2 );
      //calculate absolute error
      MUL_UP( ores->absprec, 0.5, tmp1 );
#else //C-XSC Version
      tmp1 = cxsc::_double( cxsc::subup( supMax, infMax ) );
      //calculate relative error
      tmp2 = cxsc::_double( cxsc::adddown( std::abs( supMax ), std::abs( infMax ) ) );
      ores->relprec = cxsc::_double( cxsc::divup( tmp1, tmp2 ) ); 
      //calculate absolute error
      ores->absprec = cxsc::_double( cxsc::multup( 0.5, tmp1 ) );
#endif
    }
  (ores->max) = Interval( infMax, supMax );  
  if( ores->relprec > odat->relPrec ) ores->err = PREC_ERROR;
  else                                ores->err = NO_ERROR;
}

void OptimAlgB::run()
{
  try
    {
      if( odat->maxTime != 0.0 ) QObject::startTimer( int( odat->maxTime )*1000  );      

      CPU_Time.start();
      
      Interval X( 0.0, sup( 2*PI() ) );
      
      std::vector<double> sups( odat->maxSegments );

      ores->subtlety = diam(X) / int(odat->maxSegments);
      ores->segments_sum = ores->candidates = odat->maxSegments;
      
      double 
	iX( inf(X) ), 
	sX( sup(X) );

      int ind_out( odat->maxSegments / odat->minSegments );
      
      /*---------------------------------------+
       | Screen list & enclose function values |
       +---------------------------------------*/
      Interval phi, Wf_phi, Wf_mid_phi;
      for(int k = 1; k < odat->maxSegments; k++) 
	{
	  phi = Interval( iX + (k-1) * ores->subtlety , iX + k * ores->subtlety );
	  (*(odat->mth))( phi, Wf_phi, Wf_mid_phi );
	  
	  sups[ k-1 ] = sup( Wf_phi );
	  
	  infMax = ( infMax < inf( Wf_mid_phi ) ) ? inf( Wf_mid_phi ) : infMax;
	  supMax = ( supMax < sup( Wf_phi     ) ) ? sup( Wf_phi     ) : supMax; 

	  if( shd->readStatus() == Thread::STOP || shd->readStatus() == Thread::TIMEOVER ) return;
	  
	  if( garbageflag )
	    {
	      //Garbage output
	      if( k%ind_out == 0 || k == 1 )
		{
		  QString tmp( "Index k: "
			       + QString::number(k)
			       + "\nZ = "
			       + interval2QStr( phi, 10 )
			       + "\nf(Z)-T(Z) = "
			       + interval2QStr( Wf_phi, 10 )
			       + "\nf(midZ)-T(midZ) = "
			       + interval2QStr( Wf_mid_phi, 10 )
			       + "\n\n" );
		  
		  shd->appendToGarbage( tmp );
		}
	    }

	  if( shd->readStatus() == Thread::STOP || shd->readStatus() == Thread::TIMEOVER ) return;
	}
      phi = Interval( iX + (int(odat->maxSegments) - 1) * ores->subtlety, sX );

      (*(odat->mth))( phi, Wf_phi, Wf_mid_phi);

      sups[ odat->maxSegments - 1 ] = sup( Wf_phi );
      
      infMax = ( infMax < inf( Wf_mid_phi ) ) ? inf( Wf_mid_phi ) : infMax;
      supMax = ( supMax < sup( Wf_phi     ) ) ? sup( Wf_phi     ) : supMax; 

      result_available = true;
      
      if( garbageflag )
	{
	  QString tmp( "Index k: "
		       + QString::number(odat->maxSegments - 1)
		       + "\nZ = "
		       + interval2QStr( phi, 10 )
		       + "\nf(Z)-T(Z) = "
		       + interval2QStr( Wf_phi, 10 )
		       + "\nf(midZ)-T(midZ) = "
		       + interval2QStr( Wf_mid_phi, 10 )
		       + "\n\n" );
		  
	  shd->appendToGarbage( tmp );
	}
      
      for(int k = 0; k < odat->maxSegments; k++)
	{
	  //Count promising subIntervals
	  if( sups[k] < infMax ) 
	    { 
	      (ores->candidates)--;
	    }

	  if( shd->readStatus() == Thread::STOP ) return;
	  if( shd->readStatus() == Thread::TIMEOVER )
	    {
	      collect_results();
	      return;
	    }
	}

      /*-----------------+
       | Collect results |
       +-----------------*/
      CPU_Time.stop();
      collect_results();
    }
  catch(division_by_zero)
    {
      QString mess("Division by zero occured in the calculation.");
      shd->setError( mess );
    }
  catch(function_not_defined)
    {
      QString mess("f is not analytic on the complete domain given.");
      shd->setError( mess );
    }
  catch(...)
    {
      QString mess("An unknown error occured in the calculation.");
      shd->setError( mess );
    }
}

/*

  End of File: optim.cpp

*/
